@@ -252,6 +252,14 @@ static void init_ssl(CLI *c) {
         c->error=1;
         return;
     }
+    /* Set blinding iff it's not built into our OpenSSL version */
+#if SSLEAY_VERSION_NUMBER < 0x0090702fL
+    if (options.option&OPT_CERT) 
+    	set_rsa_blinding(c->ssl);
+#else
+    log(LOG_DEBUG, "Relying on OpenSSL RSA Blinding.");
+#endif
+
 #if SSLEAY_VERSION_NUMBER >= 0x0922
     SSL_set_session_id_context(c->ssl, sid_ctx, strlen(sid_ctx));
 #endif
@@ -591,19 +599,19 @@ static int auth_libwrap(CLI *c) {
     struct request_info request;
     int result;
 
-    enter_critical_section(CRIT_LIBWRAP); /* libwrap is not mt-safe */
+    enter_critical_section(CRIT_NTOA); /* libwrap is not mt-safe, and uses
+					 inet_ntoa internally, so wrap thusly */
     request_init(&request, RQ_DAEMON, options.servname, RQ_FILE, c->local_rfd, 0);
     fromhost(&request);
     result=hosts_access(&request);
-    leave_critical_section(CRIT_LIBWRAP);
     if (!result) {
-        enter_critical_section(CRIT_NTOA); /* inet_ntoa is not mt-safe */
         log(LOG_WARNING, "Connection from %s:%d REFUSED by libwrap",
             inet_ntoa(c->addr.sin_addr), ntohs(c->addr.sin_port));
         leave_critical_section(CRIT_NTOA);
         log(LOG_DEBUG, "See hosts_access(5) for details");
         return -1; /* FAILED */
     }
+    leave_critical_section(CRIT_NTOA);
 #endif
     return 0; /* OK */
 }
@@ -686,6 +694,9 @@ static int connect_local(CLI *c) { /* spawn local process */
     char env[3][STRLEN], name[STRLEN];
     int fd[2];
     X509 *peer;
+#ifdef HAVE_PTHREAD_SIGMASK
+    sigset_t newmask;
+#endif
 
     if (options.option & OPT_PTY) {
         char tty[STRLEN];
@@ -711,16 +722,17 @@ static int connect_local(CLI *c) { /* spawn local process */
         if(!options.foreground)
             dup2(fd[1], 2);
         closesocket(fd[1]);
+        signal(SIGCHLD, SIG_DFL);
+        safecopy(env[0], "REMOTE_HOST=");
+        enter_critical_section(CRIT_NTOA); /* inet_ntoa is not mt-safe */
+        safeconcat(env[0], inet_ntoa(c->addr.sin_addr));
+        leave_critical_section(CRIT_NTOA);
+        putenv(env[0]);
         if(c->ip) {
             putenv("LD_PRELOAD=" libdir "/stunnel.so");
             /* For Tru64 _RLD_LIST is used instead */
             putenv("_RLD_LIST=" libdir "/stunnel.so:DEFAULT");
             addr.s_addr = c->ip;
-            safecopy(env[0], "REMOTE_HOST=");
-            enter_critical_section(CRIT_NTOA); /* inet_ntoa is not mt-safe */
-            safeconcat(env[0], inet_ntoa(addr));
-            leave_critical_section(CRIT_NTOA);
-            putenv(env[0]);
         }
         if(c->ssl) {
             peer=SSL_get_peer_certificate(c->ssl);
@@ -738,6 +750,10 @@ static int connect_local(CLI *c) { /* spawn local process */
                 X509_free(peer);
             }
         }
+#ifdef HAVE_PTHREAD_SIGMASK
+	sigemptyset(&newmask);
+	sigprocmask(SIG_SETMASK, &newmask, NULL);
+#endif
         execvp(options.execname, options.execargs);
         ioerror(options.execname); /* execv failed */
         _exit(1);
@@ -940,4 +956,36 @@ static void reset(int fd, char *txt) {
         log_error(LOG_DEBUG, get_last_socket_error(), txt);
 }
 
+int set_rsa_blinding(SSL *ssl) {
+#ifndef NO_RSA
+
+    /* Turn on blinding iff using RSA */
+
+    RSA *rsa;
+    EVP_PKEY *pkey;
+
+    if ( (pkey = SSL_get_privatekey(ssl)) ) {
+      if ( (rsa = EVP_PKEY_get1_RSA(pkey)) ) {
+          if ( RSA_blinding_on(rsa,NULL) ) {
+              log(LOG_DEBUG, "Stunnel manual RSA blinding enabled");
+          } else {
+              log(LOG_ERR, "Unable to set RSA blinding");
+              sslerror("RSA_blinding_on");
+              exit(1);
+          }
+          /* EVP_PKEY_get1_RSA ups the count for rsa - free extra */
+          RSA_free(rsa);
+      } else {
+          log(LOG_DEBUG, "Private key is not RSA, no blinding needed");
+      }
+    } else {
+      log(LOG_ERR, "Unable to get access to the SSL private key.");
+      sslerror("SSL_get_privatekey");
+      exit(1);
+    }
+#endif
+    return(1);
+}
+
+
 /* End of client.c */
