@@ -53,7 +53,7 @@ void log_open() { /* Win32 version */
     if(options.output_file)
         outfile=fopen(options.output_file, "a");
     if(outfile)
-        return; /* It was possible o open a log file */
+        return; /* It was possible to open a log file */
     /* TODO: Register NT EventLog source here */
     /* evt=RegisterEventSource(NULL, "stunnel"); */
     if(options.output_file)
@@ -72,10 +72,18 @@ void log_close() {
 #else /* USE_WIN32 */
 
 void log_open() { /* Unix version */
-    if(options.output_file)
-        outfile=fopen(options.output_file, "a");
-    if(outfile)
-        return; /* It was possible o open a log file */
+    if(options.output_file) {
+	int fd;
+	fd=open(options.output_file, O_CREAT|O_WRONLY|O_APPEND, 0640);
+	if (fd>=0) { /* file opened or created */
+#ifdef FD_CLOEXEC
+            fcntl(fd, F_SETFD, FD_CLOEXEC);
+#endif
+            outfile=fdopen(fd, "a");
+            if(outfile)
+                return; /* It was possible to open a log file */
+	}
+    }
     if(!options.foreground) {
 #ifdef __ultrix__
         openlog("stunnel", LOG_PID);
