@@ -180,6 +180,11 @@ unsigned long thread_id() {
     return 0L;
 }
 
+extern int signal_pipe[2];
+static void null_handler(int signum) {
+    signal(SIGCHLD, null_handler);
+}
+
 int create_client(int ls, int s, void *(*cli)(void *)) {
     switch(fork()) {
     case -1:    /* error */
@@ -187,7 +192,9 @@ int create_client(int ls, int s, void *(*cli)(void *)) {
         return -1;
     case  0:    /* child */
         closesocket(ls);
-        signal(SIGCHLD, local_handler);
+        signal(SIGCHLD, null_handler);
+        close(signal_pipe[0]);
+        close(signal_pipe[1]);
         cli((void *)s);
         exit(0);
     default:    /* parent */
