@@ -3,10 +3,12 @@
  *   Copyright (c) 1998-2001 Michal Trojnara <Michal.Trojnara@mirt.net>
  *                 All Rights Reserved
  *
- *   Version:      3.22             (stunnel.c)
- *   Date:         2001.12.20
+ *   Version:      3.26             (stunnel.c)
+ *   Date:         2003.08.29
  *
  *   Author:       Michal Trojnara  <Michal.Trojnara@mirt.net>
+ *   Maintainer:   Brian Hatch <bri@stunnel.org>
+ *                 See CREDITS for list of folks who have supplied patches.
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -48,27 +50,32 @@ static struct WSAData wsa_state;
 #endif
 
     /* Prototypes */
-static void daemon_loop();
+static void daemon_loop(void);
 #ifndef USE_WIN32
-static void daemonize();
-static void create_pid();
-static void delete_pid();
+static void daemonize(void);
+static void create_pid(void);
+static void delete_pid(void);
 #endif
 
     /* Socket functions */
-static int listen_local();
+static int listen_local(void);
 
     /* Error/exceptions handling functions */
 void ioerror(char *);
 void sockerror(char *);
 void log_error(int, int, char *);
 static char *my_strerror(int);
+#ifdef USE_PTHREAD
+static void exec_status(void);
+#endif
 #ifdef USE_FORK
-static void sigchld_handler(int);
+static void client_status(void);
 #endif
 #ifndef USE_WIN32
-void local_handler(int);
+static void sigchld_handler(int);
 static void signal_handler(int);
+static int signal_pipe[2];
+static char signal_buffer[16];
 #else
 static BOOL CtrlHandler(DWORD);
 #endif
@@ -139,6 +146,14 @@ int main(int argc, char* argv[]) { /* execution begins here 8-) */
     log(LOG_NOTICE, "%s", stunnel_info());
     if(options.option & OPT_DAEMON) { /* daemon mode */
 #ifndef USE_WIN32
+        if(pipe(signal_pipe)) {
+            ioerror("pipe");
+            exit(1);
+        }
+#ifdef FD_CLOEXEC
+	fcntl(signal_pipe[0], F_SETFD, FD_CLOEXEC);
+	fcntl(signal_pipe[1], F_SETFD, FD_CLOEXEC);
+#endif
         if(!(options.option & OPT_FOREGROUND))
             daemonize();
         create_pid();
@@ -160,11 +175,14 @@ int main(int argc, char* argv[]) { /* execution begins here 8-) */
     return 0; /* success */
 }
 
-static void daemon_loop() {
+static void daemon_loop(void) {
     int ls, s;
     struct sockaddr_in addr;
     int addrlen;
     int max_clients, fds_ulimit=-1;
+    int ready;
+    int old_val;
+    fd_set read_fds;
 
 #if defined HAVE_SYSCONF
     fds_ulimit=sysconf(_SC_OPEN_MAX);
@@ -188,25 +206,57 @@ static void daemon_loop() {
         log(LOG_ERR, "Memory allocation failed");
         exit(1);
     }
-    max_clients=max_fds>=256 ? max_fds*125/256 : (max_fds-6)/2;
+    max_clients=max_fds>=256 ? max_fds*125/256 : (max_fds-8)/2;
     log(LOG_NOTICE, "FD_SETSIZE=%d, file ulimit=%d%s -> %d clients allowed",
         FD_SETSIZE, fds_ulimit, fds_ulimit<0?" (unlimited)":"", max_clients);
     ls=listen_local();
     options.clients=0;
-#ifdef USE_FORK
+#ifndef USE_WIN32
     /* Handle signals about dead children */
     signal(SIGCHLD, sigchld_handler);
 #endif /* defined USE_FORK */
-#ifdef USE_PTHREAD
-    /* Handle signals about dead local processes */
-    signal(SIGCHLD, local_handler);
-#endif /* defined USE_PTHREAD */
     while(1) {
         addrlen=sizeof(addr);
         do {
-            s=accept(ls, (struct sockaddr *)&addr, &addrlen);
-        } while(s<0 && get_last_socket_error()==EINTR);
-        if(s<0) {
+            FD_ZERO(&read_fds);
+            FD_SET(ls, &read_fds);
+#ifndef USE_WIN32
+            FD_SET(signal_pipe[0], &read_fds);
+            ready=select((ls > signal_pipe[0]) ? ls + 1 : signal_pipe[0] + 1,
+                         &read_fds, NULL, NULL, NULL);
+#else
+            ready=select(ls + 1, &read_fds, NULL, NULL, NULL);
+#endif
+        } while(ready<=0 && get_last_error()==EINTR);
+#ifndef USE_WIN32
+        if(FD_ISSET(signal_pipe[0], &read_fds)) {
+            read(signal_pipe[0], signal_buffer, sizeof(signal_buffer));
+#ifdef	USE_PTHREAD 
+            exec_status(); /* Report status of 'exec' process */
+#else
+            client_status(); /* Report the status of the child process */
+#endif
+        }
+#endif
+        /* if we didn't also have a new connection, go back to waiting */
+        if(!FD_ISSET(ls, &read_fds)) {
+            continue;
+        }
+        /* if we also have a new connection, process it. First, put
+	 * the socket into nonblocking mode so that network errors
+	 * won't hang the next call "forever". */
+#ifndef USE_WIN32
+	old_val = fcntl(ls, F_GETFL, 0);
+	if (old_val >= 0)
+	    fcntl(ls, F_SETFL, old_val | O_NONBLOCK);
+#endif
+        s=accept(ls, (struct sockaddr *)&addr, &addrlen);
+#ifndef USE_WIN32
+	if (old_val >= 0)
+	    fcntl(ls, F_SETFL, old_val);
+#endif
+
+	if(s<0) {
             sockerror("accept");
             continue;
         }
@@ -270,7 +320,7 @@ static void daemonize() { /* go to background */
 #endif
 }
 
-static void create_pid() {
+static void create_pid(void) {
     int pf;
     char pid[STRLEN];
     struct stat sb;
@@ -335,7 +385,7 @@ static void create_pid() {
     atexit(delete_pid);
 }
 
-static void delete_pid() {
+static void delete_pid(void) {
     log(LOG_DEBUG, "removing pid file %s", options.pidfile);
     if((unsigned long)getpid()!=options.dpid)
         return; /* Current process is not main daemon process */
@@ -344,7 +394,7 @@ static void delete_pid() {
 }
 #endif /* defined USE_WIN32 */
 
-static int listen_local() { /* bind and listen on local interface */
+static int listen_local(void) { /* bind and listen on local interface */
     struct sockaddr_in addr;
     int ls;
 
@@ -370,7 +420,10 @@ static int listen_local() { /* bind and listen on local interface */
         sockerror("listen");
         exit(1);
     }
-
+#ifdef FD_CLOEXEC
+    fcntl(ls, F_SETFD, FD_CLOEXEC); /* close socket in child execvp */
+#endif
+	    
 #ifndef USE_WIN32
     if(options.setgid_group) {
         struct group *gr;
@@ -575,58 +628,67 @@ static char *my_strerror(int errnum) {
 }
 
 #ifdef USE_FORK
-static void sigchld_handler(int sig) { /* Dead children detected */
+static void client_status(void) { /* dead children detected */
     int pid, status;
 
 #ifdef HAVE_WAIT_FOR_PID
     while((pid=wait_for_pid(-1, &status, WNOHANG))>0) {
-        options.clients--; /* One client less */
+        option.clients--; /* one client less */
 #else
-    if((pid=wait(&status))>0) {
-        options.clients--; /* One client less */
+        if((pid=wait(&status))>0) {
+            option.clients--; /* one client less */
 #endif
 #ifdef WIFSIGNALED
         if(WIFSIGNALED(status)) {
-            log(LOG_DEBUG, "%s[%d] terminated on signal %d (%d left)",
-                options.servname, pid, WTERMSIG(status), options.clients);
+            log(LOG_DEBUG, "Process %d terminated on signal %d (%d left)",
+                pid, WTERMSIG(status), num_clients);
         } else {
-            log(LOG_DEBUG, "%s[%d] finished with code %d (%d left)",
-                options.servname, pid, WEXITSTATUS(status), options.clients);
+            log(LOG_DEBUG, "Process %d finished with code %d (%d left)",
+                pid, WEXITSTATUS(status), num_clients);
         }
     }
 #else
-        log(LOG_DEBUG, "%s[%d] finished with code %d (%d left)",
-            options.servname, pid, status, options.clients);
+        log(LOG_DEBUG, "Process %d finished with code %d (%d left)",
+            pid, status, num_clients);
     }
 #endif
-    signal(SIGCHLD, sigchld_handler);
 }
 #endif
 
-#ifndef USE_WIN32
 
-void local_handler(int sig) { /* Dead of local (-l) process detected */
+#ifdef USE_PTHREAD
+static void exec_status(void) { /* dead local ('exec') process detected */
     int pid, status;
 
 #ifdef HAVE_WAIT_FOR_PID
     while((pid=wait_for_pid(-1, &status, WNOHANG))>0) {
 #else
-    if((pid=wait(&status))>0) {
+        if((pid=wait(&status))>0) {
 #endif
 #ifdef WIFSIGNALED
-        if(WIFSIGNALED(status)) {
-            log(LOG_DEBUG, "Local process %s (PID=%lu) terminated on signal %d",
-                options.servname, pid, WTERMSIG(status));
-        } else {
-            log(LOG_DEBUG, "Local process %s (PID=%lu) finished with code %d",
-                options.servname, pid, WEXITSTATUS(status));
-        }
+             if(WIFSIGNALED(status)) {
+                 log(LOG_INFO, "Local process %d terminated on signal %d",
+                     pid, WTERMSIG(status));
+             } else {
+                 log(LOG_INFO, "Local process %d finished with code %d",
+                     pid, WEXITSTATUS(status));
+             }
 #else
-        log(LOG_DEBUG, "Local process %s (PID=%lu) finished with status %d",
-            options.servname, pid, status);
+        log(LOG_INFO, "Local process %d finished with status %d",
+            pid, status);
 #endif
-    }
-    signal(SIGCHLD, local_handler);
+        }
+}
+#endif /* !defined USE_WIN32 */
+
+
+#ifndef USE_WIN32
+static void sigchld_handler(int sig) { /* Death of child process detected */
+    int save_errno=errno;
+    
+    write(signal_pipe[1], signal_buffer, 1);
+    signal(SIGCHLD, sigchld_handler);
+    errno=save_errno;
 }
 
 static void signal_handler(int sig) { /* Signal handler */
@@ -634,7 +696,7 @@ static void signal_handler(int sig) { /* Signal handler */
     exit(3);
 }
 
-#else /* !defined USE_win32 */
+#else /* !defined USE_WIN32 */
 
 static BOOL CtrlHandler(DWORD fdwCtrlType) {
     switch(fdwCtrlType) {
